"""
Classes to add request.breadcrumbs as one class to have a list of breadcrumbs
TODO: maybe is better to move to contrib/breadcrumbs
"""

from django.conf import settings
from django.utils.translation import ugettext as _
from django.utils.safestring import mark_safe


class Singleton(object):
    """
    We use a simple singleton pattern in Breadcrumbs.
    Example from http://svn.ademar.org/code/trunk/junk-code/singleton_vs_borg.py
    """
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it._1st_init(*args, **kwds)
        return it

    def _1st_init(self, *args, **kwds):
        pass


class BreadcrumbsInvalidFormat(Exception):
    """
    Simple exception that can be extended
    """
    pass


class BreadcrumbsNotSet(Exception):
    """
    Raised in utils.breadcrumbs_for_flatpages when we not have breadcrumbs in
    request.
    """
    pass


class Breadcrumb(object):
    """
    Breadcrumb can have methods to customize breadcrumb object, Breadcrumbs
    class send to us name and url.
    """
    def __init__(self, name, url):
        # HERE
        #
        # If I don't use force_unicode, always runs ok, but have problems on
        # template with unicode text
        self.name = name
        self.url = url

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return u"%s,%s" % (self.name, self.url)

    def __repr__(self):
        return u"Breadcrumb <%s,%s>" % (self.name, self.url)


class Breadcrumbs(Singleton):
    """
    Breadcrumbs maintain a list of breadcrumbs that you can get interating with
    class or with get_breadcrumbs().
    """
    def _1st_init(self, *args, **kwargs):
        """
        singleton function that start some variables
        """
        self._clean()
        self.__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if not len(args) and not len(kwargs):
            return self
        return self.__init__(*args, **kwargs)

    def __fill_home(self):
        # fill home if settings.BREADCRUMBS_AUTO_HOME is True
        if self.__autohome and len(self.__bds) == 0:
            home_title = getattr(settings, 'BREADCRUMBS_HOME_TITLE', _(u'Home'))
            self.__fill_bds((home_title, u"/"))

    def _clean(self):
        self.__bds = []
        self.__autohome = getattr(settings, 'BREADCRUMBS_AUTO_HOME', False)
        self.__urls = []
        self.__fill_home()

    def __init__(self, *args, **kwargs):
        """
        Call validate and if ok, call fill bd
        """
        # match Breadcrumbs( 'name', 'url' )
        if len(args) == 2 and type(args[0]) not in (list, tuple):
            if(self.__validate(args, 0)):
                self.__fill_bds(args)
        # match ( ( 'name', 'url'), ..) and samething with list
        elif len(args) == 1 and type(args[0]) in (list, tuple) \
                and len(args[0]) > 0:
            for i, arg in enumerate(args[0]):
                if self.__validate(arg, i):
                    self.__fill_bds(arg)
        # try to ( obj1, obj2, ... ) and samething with list
        else:
            for i, arg in enumerate(args):
                if(self.__validate(arg, i)):
                    self.__fill_bds(arg)

    def __validate(self, obj, index):
        """
        check for object type and return a string as name for each item of a
        list or tuple with items, if error was found raise
        BreadcrumbsInvalidFormat
        """
        # for list or tuple
        if type(obj) in (list, tuple):
            if len(obj) == 2:
                if (not obj[0] and not obj[1]) or \
                        (type(obj[0]) not in (str, unicode) and \
                        type(obj[1]) not in (str, unicode)):
                    raise BreadcrumbsInvalidFormat(u"Invalid format for \
                        breadcrumb %s in %s" % (index, type(obj).__name__))
            if len(obj) != 2:
                raise BreadcrumbsInvalidFormat(
                    u"Wrong itens number in breadcrumb %s in %s. \
                    You need to send as example (name,url)" % \
                    (index, type(obj).__name__)
                )
        # for objects
        elif not hasattr(obj, 'name') and not hasattr(obj, 'url'):
            raise BreadcrumbsInvalidFormat(u"You need to use a tuple like "
                "(name,url) or object with name and url attributes for "
                "breadcrumb.")
        return True

    def __fill_bds(self, bd):
        """
        simple interface to add Breadcrumb to bds
        """
        if hasattr(bd, 'name') and hasattr(bd, 'url'):
            bd = Breadcrumb(bd.name, bd.url)
        else:
            bd = Breadcrumb(*bd)
        if bd.url not in self.__urls:
            self.__bds.append(bd)
            self.__urls.append(bd.url)

    def __iter__(self):
        return iter(self.__bds)

    def __getitem__(self, key):
        return self.__bds[key]

    def __repr__(self):
        return self.__unicode__()

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return u"Breadcrumbs <%s>" % u", ".join([mark_safe(item.name) for item \
                                                    in self[:10]] + [u' ...'])

    def all(self):
        return self.__bds

# vim: ts=4 sts=4 sw=4 et ai



from django.conf import settings
from django.contrib.flatpages.views import render_flatpage
from django.http import Http404, HttpResponsePermanentRedirect
from utils import breadcrumbs_for_flatpages, get_flapage_from_cache


def flatpage(request, url):
    """
    Public interface to the flat page view.

    Models: `flatpages.flatpages`
    Templates: Uses the template defined by the ``template_name`` field,
        or `flatpages/default.html` if template_name is not defined.
    Context:
        flatpage
            `flatpages.flatpages` object
    """

    if not url.startswith('/'):
        url = '/' + url
    try:
        # try load flatpage from cache, else, update cache and get from DB
        f = get_flapage_from_cache(url)
    except Http404:
        if not url.endswith('/') and settings.APPEND_SLASH:
            url += '/'
            f = get_flapage_from_cache(url)
            return HttpResponsePermanentRedirect('%s/' % request.path)
        else:
            raise

    # create breadcrumbs
    breadcrumbs_for_flatpages(request, f)

    return render_flatpage(request, f)



# -*- coding: utf-8 -*-
from django.contrib.flatpages.models import FlatPage
from django.http import Http404
from breadcrumbs import Breadcrumbs, BreadcrumbsNotSet
from django.conf import settings
from django.core.cache import cache


def make_flatpages_cache_key():
    """
    Create a cache key based on some basic data, respecting defined site.
    """
    key = "flatpages_cache_%s-%s" % (hash(settings.SITE_ID),
                                      hash(settings.SECRET_KEY))

    return key


def get_flapage_from_cache(url):
    """
    Try get flatpage from cache entry with all flatpages by url.
    If not found, create cache and return flatpage from db.

    This probably avoid some hits on DB.
    """
    site_id = settings.SITE_ID
    cache_key = make_flatpages_cache_key()
    flatpages = cache.get(cache_key)
    if flatpages and url in flatpages:
        return flatpages[url]

    # flatpages cache not exist or flatpage not found.

    # 1. get all flatpages.
    flatpages = dict([(f.url, f) for f in
        FlatPage.objects.filter(sites__id__exact=site_id).order_by('url')])

    # 2. if url not in flatpages, raise Http404
    if url not in flatpages:
        raise Http404

    # 3. if url in flatpages, recreate cache and return flatpage
    cache.delete(cache_key)
    cache.add(cache_key, flatpages)
    return flatpages[url]


def breadcrumbs_for_flatpages(request, flatpage):
    """ given request and flatpage instance create breadcrumbs for all flat
    pages """
    if not hasattr(request, 'breadcrumbs') or \
                            not isinstance(request.breadcrumbs, Breadcrumbs):
        raise BreadcrumbsNotSet(u"You need to setup breadcrumbs to use this "
                                u"function.")

    if not isinstance(flatpage, FlatPage) or not hasattr(flatpage, 'id'):
        raise TypeError(u"flatpage argument isn't a FlatPage instance or not "
                        u"have id.")

    # URL for a flatpage can be composed of other flatpages, ex:
    #
    # We have:
    #   flatpage01 = /flat01/
    #   flatpage02 = /flat01/flat02/
    #   flatpage03 = /flat01/flat02/flat03/
    #
    # In breadcrumbs we want to know each title of each page, so we split url
    # in parts, and try to get flatpage title.
    #
    # However, you can define something like that in your urls.py:
    #   (r'^pages/', include('breadcrumbs.urls')),
    # And, we will never know what is /pages/, so we ignore it for now.
    paths = []
    for part in request.path_info.split(u"/"):
        # When split we have u"" for slashes
        if len(part) == 0:
            continue
        # Add slash agai
        if not part.startswith(u"/"):
            part = u"/" + part
        if not part.endswith(u"/"):
            part = part + u"/"
        # If we have something on paths, url for flatpage is composed of what we
        # have in path + part. Note that strings in path not have last slash, but
        # part have.
        if len(paths) > 0:
            url = u"".join(paths + [part])
        else:
            url = part
        # if part of url is same url of flatpage instance, we don't hit
        # database again, we get page from FlatPage instance.
        # If part of url isn't same url of flatpage instance, we try to get it.
        # If page doesn't exist, we just continue to next part.
        if url == flatpage.url:
            request.breadcrumbs(flatpage.title, flatpage.url)
        else:
            try:
                f = FlatPage.objects.get(url=url)
            except FlatPage.DoesNotExist:
                # TODO: this part can be a view, maybe is a good idea get that
                # view and check for viewfunc.breadcrumb_title or
                # viewclass.breadcrumb_title attributes.
                continue
            else:
                request.breadcrumbs(f.title, f.url)
        # add last part of path in paths with one slash
        paths.append(u"/" + url[1:-1].rpartition(u"/")[-1])



# -*- coding: utf-8 -*-

# TODO: Write unit tests!!!

# run this script on your shell
from breadcrumbs import Breadcrumbs

bds = Breadcrumbs()

# fill one per time
for i in range(5):
    bds('name%s' % i, 'url%s' % i)


# create a simple class to emulate one object with name and url
class EmulateObj(object):
    def __init__(self, *args):
        self.name = args[0]
        self.url = args[1]

# add 10 objects
bds([EmulateObj('name %s' % (i + 10), 'url %s' % (i + 10)) for i in range(10)])

# print all
for bd in bds:
    print bd.name, bd.url



# coding: utf-8

import os
from django.conf import settings
from django.test import TestCase

from breadcrumbs.breadcrumbs import (Breadcrumb, Breadcrumbs,
    BreadcrumbsInvalidFormat, BreadcrumbsNotSet)

class BreadcrumbsTest(TestCase):

    def setUp(self):
        self.old_MIDDLEWARE_CLASSES = settings.MIDDLEWARE_CLASSES
        breadcrumbs_middleware_class = 'breadcrumbs.middleware.BreadcrumbsMiddleware'
        if breadcrumbs_middleware_class not in settings.MIDDLEWARE_CLASSES:
            settings.MIDDLEWARE_CLASSES += (breadcrumbs_middleware_class,)
        self.old_TEMPLATE_DIRS = settings.TEMPLATE_DIRS
        settings.TEMPLATE_DIRS = (
            os.path.join(
                os.path.dirname(__file__),
                'templates'
            ),
        )
        # self.old_LOGIN_URL = settings.LOGIN_URL
        # settings.LOGIN_URL = '/accounts/login/'

        # now we start singleton
        self.breadcrumbs = Breadcrumbs()

    def tearDown(self):
        settings.MIDDLEWARE_CLASSES = self.old_MIDDLEWARE_CLASSES
        settings.TEMPLATE_DIRS = self.old_TEMPLATE_DIRS
        # settings.LOGIN_URL = self.old_LOGIN_URL

        # kill singleton
        Breadcrumbs('a','/a')
        Breadcrumbs('b','/b')
        print Breadcrumbs()
        self.breadcrumbs._drop_it()
        del(self.breadcrumbs)
        Breadcrumbs('c','/c')
        print Breadcrumbs()

    def test_breadcrumb_class(self):
        b = Breadcrumb("Home", "/")
        self.assertEqual(b.name, "Home")
        self.assertEqual(b.url, "/")

    def test_breadcrumbs_singleton(self):





# coding: utf-8

instance = None


class Bread(object):

    def __init__(self, *a, **kw):
        import ipdb; ipdb.set_trace()
        print instance

    def set_instance(self, v):
        instance = v



from setuptools import setup, find_packages
setup(
    name="django-breadcrumbs",
    version="1.1.2",
    packages=find_packages(exclude=('breadcrumbs_sample*', 'sample_d14*')),
    license='LICENSE',
    description="Easy to use generic breadcrumbs system for Django framework.",
    author="Felipe 'chronos' Prenholato",
    author_email="philipe.rp@gmail.com",
    url="http://github.com/chronossc/django-breadcrumbs",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    install_requires=["Django>=1.2.7"],
)



# -*- coding: utf-8 -*-
from django.conf import settings
from django.http import Http404
from breadcrumbs import Breadcrumbs
from views import flatpage


class BreadcrumbsMiddleware(object):

    def process_request(self, request):
        request.breadcrumbs = Breadcrumbs()
        request.breadcrumbs._clean()


class FlatpageFallbackMiddleware(object):
    def process_response(self, request, response):
        # do nothing if flatpages middleware isn't enabled, also if response
        # code isn't 404.
        if response.status_code != 404:
            return response
        try:
            return flatpage(request, request.path_info)
        # Return the original response if any errors happened. Because this
        # is a middleware, we can't assume the errors will be caught elsewhere.
        except Http404:
            return response
        except:
            if settings.DEBUG:
                raise
            return response



# -*- coding: utf-8 -*-
from django.contrib.flatpages.models import FlatPage
from django.core.cache import cache
from django.db.models.signals import post_save
from utils import make_flatpages_cache_key


def clean_flatpages_cache(sender, **kw):
    """
    Invalidate flatpages cache, because some flatpage was saved!
    """
    cache.delete(make_flatpages_cache_key())

post_save.connect(clean_flatpages_cache, sender=FlatPage)
